-------------------------- MODULE LeafSet ------------------------------------------

EXTENDS Ring

LeafSet ==
  {ls \in [node: I, left: SUBSET I, right: SUBSET I]: 
    /\ ls.node \notin ls.left
    /\ ls.node \notin ls.right
    /\ Cardinality(ls.left) =< L
    /\ Cardinality(ls.right) =< L}

LeafSetContent(ls) == ls.left \cup ls.right \cup {ls.node}

EmptyLS(i) == [node |-> i, left|-> {}, right|-> {}]

LeftNeighbor(ls) == 
    ClosestFromTheLeft(ls.node, ls.left)

RightNeighbor(ls) == 
    ClosestFromTheRight(ls.node, ls.right)

LeftmostNode(ls) ==
    FarthestFromTheLeft(ls.node, ls.left)

RightmostNode(ls) ==
    FarthestFromTheRight(ls.node, ls.right)

AddToLS(delta, ls) ==
(* Adds the nodes in delta to the leaf set ls. *)
  LET 
    i == ls.node
    can == (ls.left \cup ls.right \cup delta) \ {i} 
    newleft == 
      ClosestNodesFromTheLeft(i, can, L)
    newright == 
      ClosestNodesFromTheRight(i, can, L)
  IN [node |-> i, left |-> newleft, right |-> newright] 
  
RemoveFromLS(delta, ls) ==  
(* Removes the nodes in delta from the leaf set ls. *)
  LET 
    i == ls.node
    can == (ls.left \cup ls.right) \ ({i} \cup delta)
    newleft == 
      ClosestNodesFromTheLeft(i, can, L)
    newright == 
      ClosestNodesFromTheRight(i, can, L)
  IN [node |-> i, left |-> newleft, right |-> newright] 
  
LeftCoverage(ls) ==
(* The left bound of ls.node's coverage/key ownership, excluding edge node. *)
  IF LeftNeighbor(ls) = ls.node
  THEN ls.node
  ELSE (LeftNeighbor(ls) + (ClockwiseDistance(LeftNeighbor(ls), ls.node) \div 2 + 1)) % RingSize
  
RightCoverage(ls) ==
(* The right bound of ls.node's coverage/key ownership, including edge node. *)
    IF RightNeighbor(ls) = ls.node
    THEN (RingSize + ls.node - 1) % RingSize
    ELSE (ls.node + ClockwiseDistance(ls.node, RightNeighbor(ls)) \div 2) % RingSize

Covers(ls, k) == ClockwiseArc(LeftCoverage(ls), k, RightCoverage(ls))

Overlaps(ls) == ls.left \cap ls.right # {}

IsComplete(ls) ==
  /\ Cardinality(ls.left) = L 
  /\ Cardinality(ls.right) = L

IsBalanced(ls) ==
  Cardinality(ls.left) = Cardinality(ls.right)
        
IsProper(ls) ==
  /\ \A x \in ls.left \ ls.right, y \in ls.right:
      ClockwiseArc(y, x, ls.node)
  /\ \A x \in ls.right \ ls.left, y \in ls.left:
      ClockwiseArc(ls.node, x, y)

IsOrganized(ls) ==
    \/ ls.left \ ls.right # {}
    \/ ls.right \ ls.left # {}
  => IsComplete(ls)

IsMaximal(ls) == 
  AddToLS({}, ls) = ls
  
===============================================================================